/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check;

import de.hft.stuttgart.citydoctor2.check.error.AllPolygonsWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.ConsecutivePointSameError;
import de.hft.stuttgart.citydoctor2.check.error.DependenciesNotMetError;
import de.hft.stuttgart.citydoctor2.check.error.DistanceError;
import de.hft.stuttgart.citydoctor2.check.error.DuplicatePointError;
import de.hft.stuttgart.citydoctor2.check.error.EdgeIntersectionError;
import de.hft.stuttgart.citydoctor2.check.error.HoleOutsideError;
import de.hft.stuttgart.citydoctor2.check.error.InteriorDisconnectedError;
import de.hft.stuttgart.citydoctor2.check.error.ManifoldEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.MultipleConnectedComponentsError;
import de.hft.stuttgart.citydoctor2.check.error.NestedRingError;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldVertexError;
import de.hft.stuttgart.citydoctor2.check.error.NormalDeviationError;
import de.hft.stuttgart.citydoctor2.check.error.NotCeilingError;
import de.hft.stuttgart.citydoctor2.check.error.NotFloorError;
import de.hft.stuttgart.citydoctor2.check.error.NotGroundError;
import de.hft.stuttgart.citydoctor2.check.error.NotWallError;
import de.hft.stuttgart.citydoctor2.check.error.NullAreaError;
import de.hft.stuttgart.citydoctor2.check.error.PointTouchesEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonSelfIntError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.RingNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.RingSelfIntError;
import de.hft.stuttgart.citydoctor2.check.error.SameOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.SchematronError;
import de.hft.stuttgart.citydoctor2.check.error.SolidNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.SolidSelfIntError;
import de.hft.stuttgart.citydoctor2.check.error.SurfaceUnfragmentedError;
import de.hft.stuttgart.citydoctor2.check.error.TinyEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.TooFewPointsError;
import de.hft.stuttgart.citydoctor2.check.error.TooFewPolygonsError;
import de.hft.stuttgart.citydoctor2.check.error.UnknownCheckError;

/**
 * Visitor pattern interface used in the healing procedures. The modification
 * listener is to report changed polygons and the returned boolean indicates
 * that changes have happened.
 * 
 * @author Matthias Betz
 *
 */
public interface HealingMethod {

	default boolean visit(CheckError e, ModificationListener l) {
		return false;
	}
	
	default boolean visit(TinyEdgeError e, ModificationListener l) {
		return false;
	}

	default boolean visit(EdgeIntersectionError edgeIntersectionError, ModificationListener l) {
		return false;
	}

	default boolean visit(DuplicatePointError duplicatePointError, ModificationListener l) {
		return false;
	}

	default boolean visit(AllPolygonsWrongOrientationError err, ModificationListener l) {
		return false;
	}

	default boolean visit(DependenciesNotMetError err, ModificationListener l) {
		return false;
	}

	default boolean visit(DistanceError err, ModificationListener l) {
		return false;
	}

	default boolean visit(ConsecutivePointSameError err, ModificationListener l) {
		return false;
	}

	default boolean visit(HoleOutsideError err, ModificationListener l) {
		return false;
	}

	default boolean visit(InteriorDisconnectedError err, ModificationListener l) {
		return false;
	}

	default boolean visit(ManifoldEdgeError err, ModificationListener l) {
		return false;
	}

	default boolean visit(MultipleConnectedComponentsError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NestedRingError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NonManifoldVertexError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NormalDeviationError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NullAreaError err, ModificationListener l) {
		return false;
	}

	default boolean visit(PolygonSelfIntError err, ModificationListener l) {
		return false;
	}

	default boolean visit(PolygonWrongOrientationError err, ModificationListener l) {
		return false;
	}

	default boolean visit(RingNotClosedError err, ModificationListener l) {
		return false;
	}

	default boolean visit(RingSelfIntError err, ModificationListener l) {
		return false;
	}

	default boolean visit(SameOrientationError err, ModificationListener l) {
		return false;
	}

	default boolean visit(SolidNotClosedError err, ModificationListener l) {
		return false;
	}

	default boolean visit(SolidSelfIntError err, ModificationListener l) {
		return false;
	}

	default boolean visit(TooFewPointsError err, ModificationListener l) {
		return false;
	}

	default boolean visit(TooFewPolygonsError err, ModificationListener l) {
		return false;
	}

	default boolean visit(UnknownCheckError err, ModificationListener l) {
		return false;
	}

	default boolean visit(PointTouchesEdgeError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NotCeilingError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NotFloorError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NotWallError err, ModificationListener l) {
		return false;
	}

	default boolean visit(NotGroundError err, ModificationListener l) {
		return false;
	}
	
	default boolean visit(SchematronError err, ModificationListener l) {
		return false;
	}
	
	default boolean visit(SurfaceUnfragmentedError err, ModificationListener l) {
		return false;
	}

	public HealingMethod createNew();

}
