/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.utils.SerializablePair;

/**
 * If the exterior ring and interior rings are intersecting each other, it
 * results in a polygon self intersection and this error is created.
 * 
 * @author Matthias Betz
 *
 */
public class PolygonSelfIntError extends CheckError {

	private static final long serialVersionUID = -7043521697085991901L;
	
	private Polygon p;
	private SerializablePair<LinearRing, LinearRing> intersectingRings;

	public PolygonSelfIntError(Polygon p, SerializablePair<LinearRing, LinearRing> intersectingRings) {
		super(ErrorId.GE_P_INTERSECTING_RINGS, ErrorType.ERROR, p);
		this.p = p;
		this.intersectingRings = intersectingRings;
	}

	public Polygon getPolygon() {
		return p;
	}

	public SerializablePair<LinearRing, LinearRing> getIntersectingRings() {
		return intersectingRings;
	}

	@Override
	public void report(ErrorReport report) {
		report.add(p);
		report.add(intersectingRings.getValue0());
		report.add(intersectingRings.getValue1());
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "PolygonSelfIntError [p=" + p + ", intersectingRings=" + intersectingRings + "]";
	}
}
