/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check.error;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ErrorReport;
import de.hft.stuttgart.citydoctor2.check.ErrorType;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;

/**
 * If a polygon is wrongly oriented it results in edges being traversed in the
 * same direction. This error stores all edges that have been traversed in the
 * same direction twice.
 * 
 * @author Matthias Betz
 *
 */
public class PolygonWrongOrientationError extends CheckError {

	private static final long serialVersionUID = -3006643655816847097L;
	
	private List<Edge> edges;
	private Geometry g;

	public PolygonWrongOrientationError(Geometry g, List<Edge> edges) {
		super(ErrorId.GE_S_POLYGON_WRONG_ORIENTATION, ErrorType.ERROR, g);
		this.edges = edges;
		this.g = g;
	}

	public List<Edge> getEdges() {
		return edges;
	}

	public Geometry getGeometry() {
		return g;
	}

	@Override
	public void report(ErrorReport report) {
		report.add(g);
		Set<Polygon> possiblePolygons = new HashSet<>();
		for (Edge e : edges) {
			possiblePolygons.addAll(e.getAdjacentPolygons());
		}
		report.addPolygons("possible polygons", possiblePolygons);
	}

	@Override
	public void accept(ErrorVisitor errorVisitor) {
		errorVisitor.visit(this);
	}

	@Override
	public boolean accept(HealingMethod method, ModificationListener l) {
		return method.visit(this, l);
	}

	@Override
	public String toString() {
		return "PolygonWrongOrientationError [edges=" + edges + ", g=" + g + "]";
	}
}
