/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import java.util.ArrayList;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.factory.GMLGeometryFactory;
import org.citygml4j.model.citygml.building.BoundarySurfaceProperty;
import org.citygml4j.model.gml.geometry.aggregates.MultiSurface;
import org.citygml4j.model.gml.geometry.aggregates.MultiSurfaceProperty;
import org.citygml4j.model.gml.geometry.primitives.Solid;
import org.citygml4j.model.gml.geometry.primitives.SolidProperty;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;

/**
 * Container for attributes of an abstract building. It can be a Building or
 * BuildingPart.
 * 
 * @author Matthias Betz
 *
 */
public abstract class AbstractBuilding extends CityObject {

	private static final long serialVersionUID = -2196414503088741206L;

	private static final Logger logger = LogManager.getLogger(AbstractBuilding.class);

	private List<BuildingInstallation> buildingInstallations = new ArrayList<>(2);
	private List<BoundarySurface> boundarySurfaceList = new ArrayList<>();

	private org.citygml4j.model.citygml.building.AbstractBuilding ab;

	/**
	 * Getter for all boundary surfaces contained in this building.
	 * 
	 * @return the boundary surfaces
	 */
	public List<BoundarySurface> getBoundarySurfaces() {
		return boundarySurfaceList;
	}

	@Override
	public org.citygml4j.model.citygml.building.AbstractBuilding getGmlObject() {
		return ab;
	}

	@Override
	public FeatureType getFeatureType() {
		return FeatureType.BUILDING;
	}

	@Override
	public void unsetGmlGeometries() {
		ab.unsetLod1MultiSurface();
		ab.unsetLod2MultiSurface();
		ab.unsetLod3MultiSurface();
		ab.unsetLod4MultiSurface();
		ab.unsetLod1Solid();
		ab.unsetLod2Solid();
		ab.unsetLod3Solid();
		ab.unsetLod4Solid();
		for (BoundarySurface bs : boundarySurfaceList) {
			bs.unsetGmlGeometries();
		}
		for (BuildingInstallation bi : buildingInstallations) {
			bi.unsetGmlGeometries();
		}
	}

	@Override
	public void reCreateGeometries(GMLGeometryFactory factory, ParserConfiguration config) {
		for (Geometry geom : getGeometries()) {
			if (geom.getType() == GeometryType.MULTI_SURFACE) {
				MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
				setMultiSurfaceAccordingToLod(geom, ms);
			} else {
				Solid solid = CityGmlUtils.createSolid(geom, factory, config);
				setSolidAccordingToLod(geom, solid);
			}
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			reCreateBoundarySurface(factory, config, bs);
		}
		for (BuildingInstallation bi : buildingInstallations) {
			bi.reCreateGeometries(factory, config);
		}
	}

	private void reCreateBoundarySurface(GMLGeometryFactory factory, ParserConfiguration config, BoundarySurface bs) {
		if (bs.getGeometries().isEmpty()) {
			for (BoundarySurfaceProperty bsp : ab.getBoundedBySurface()) {
				if (bsp.getBoundarySurface() != null && bsp.getBoundarySurface() == bs.getGmlObject()) {
					logger.warn("Found empty boundary surface: {}, removing from building", bs.getGmlId());
					ab.getBoundedBySurface().remove(bsp);
					break;
				}
			}
			return;
		}
		bs.reCreateGeometries(factory, config);
	}

	private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
		switch (geom.getLod()) {
		case LOD1:
			ab.setLod1MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD2:
			ab.setLod2MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD3:
			ab.setLod3MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD4:
			ab.setLod4MultiSurface(new MultiSurfaceProperty(ms));
			break;
		default:
			throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to buildings");
		}
	}

	private void setSolidAccordingToLod(Geometry geom, Solid solid) {
		switch (geom.getLod()) {
		case LOD1:
			ab.setLod1Solid(new SolidProperty(solid));
			break;
		case LOD2:
			ab.setLod2Solid(new SolidProperty(solid));
			break;
		case LOD3:
			ab.setLod3Solid(new SolidProperty(solid));
			break;
		case LOD4:
			ab.setLod4Solid(new SolidProperty(solid));
			break;
		default:
			throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to buildings");
		}
	}

	@Override
	public void accept(Check c) {
		super.accept(c);
		if (c.canExecute(this)) {
			c.check(this);
		}
		for (BuildingInstallation bi : buildingInstallations) {
			bi.accept(c);
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			bs.accept(c);
		}
	}

	@Override
	public void collectContainedErrors(List<CheckError> errors) {
		super.collectContainedErrors(errors);
		for (BuildingInstallation bi : buildingInstallations) {
			bi.collectContainedErrors(errors);
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			bs.collectContainedErrors(errors);
		}
	}

	@Override
	public void clearAllContainedCheckResults() {
		super.clearAllContainedCheckResults();
		for (BuildingInstallation bi : buildingInstallations) {
			bi.clearAllContainedCheckResults();
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			bs.clearAllContainedCheckResults();
		}
	}

	@Override
	public boolean containsError(CheckId checkIdentifier) {
		boolean hasError = super.containsError(checkIdentifier);
		if (hasError) {
			return true;
		}
		for (BuildingInstallation bi : buildingInstallations) {
			if (bi.containsError(checkIdentifier)) {
				return true;
			}
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			if (bs.containsError(checkIdentifier)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean containsAnyError() {
		boolean hasError = super.containsAnyError();
		if (hasError) {
			return true;
		}
		for (BuildingInstallation bi : buildingInstallations) {
			if (bi.containsAnyError()) {
				return true;
			}
		}
		for (BoundarySurface bs : boundarySurfaceList) {
			if (bs.containsAnyError()) {
				return true;
			}
		}
		return false;
	}

	void setCityGmlBuilding(org.citygml4j.model.citygml.building.AbstractBuilding ab) {
		this.ab = ab;
	}

	public void addBoundarySurface(BoundarySurface bs) {
		boundarySurfaceList.add(bs);
		bs.setParent(this);
	}

	public void addBuildingInstallation(BuildingInstallation coBi) {
		buildingInstallations.add(coBi);
		coBi.setParent(this);
	}

	public void setGmlObject(org.citygml4j.model.citygml.building.AbstractBuilding ab) {
		this.ab = ab;
	}

	public List<BuildingInstallation> getBuildingInstallations() {
		return buildingInstallations;
	}
}
