/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import java.util.LinkedList;
import java.util.List;

import org.citygml4j.factory.GMLGeometryFactory;
import org.citygml4j.model.citygml.bridge.AbstractBridge;
import org.citygml4j.model.gml.geometry.aggregates.MultiSurface;
import org.citygml4j.model.gml.geometry.aggregates.MultiSurfaceProperty;
import org.citygml4j.model.gml.geometry.primitives.Solid;
import org.citygml4j.model.gml.geometry.primitives.SolidProperty;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;

/**
 * Bridge representation object of CityGML bridge objects
 * 
 * @author Matthias Betz
 *
 */
public class BridgeObject extends CityObject {

	private static final long serialVersionUID = 6301112640328373842L;

	public enum BridgeType {
		BRIDGE, BRIDGE_PART
	}

	private AbstractBridge ab;
	private BridgeType type;

	private List<BoundarySurface> boundarySurfaces = new LinkedList<>();

	public BridgeObject(BridgeType type, AbstractBridge ab) {
		this.ab = ab;
		this.type = type;
	}

	@Override
	public FeatureType getFeatureType() {
		return FeatureType.BRIDGE;
	}

	@Override
	public void reCreateGeometries(GMLGeometryFactory factory, ParserConfiguration config) {
		for (Geometry geom : getGeometries()) {
			if (geom.getType() == GeometryType.MULTI_SURFACE) {
				MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
				setMultiSurfaceAccordingToLod(geom, ms);
			} else {
				Solid solid = CityGmlUtils.createSolid(geom, factory, config);
				setSolidAccordingToLod(geom, solid);
			}
		}
	}

	private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
		switch (geom.getLod()) {
		case LOD1:
			ab.setLod1MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD2:
			ab.setLod2MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD3:
			ab.setLod3MultiSurface(new MultiSurfaceProperty(ms));
			break;
		case LOD4:
			ab.setLod4MultiSurface(new MultiSurfaceProperty(ms));
			break;
		default:
			throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to bridges");
		}
	}

	private void setSolidAccordingToLod(Geometry geom, Solid solid) {
		switch (geom.getLod()) {
		case LOD1:
			ab.setLod1Solid(new SolidProperty(solid));
			break;
		case LOD2:
			ab.setLod2Solid(new SolidProperty(solid));
			break;
		case LOD3:
			ab.setLod3Solid(new SolidProperty(solid));
			break;
		case LOD4:
			ab.setLod4Solid(new SolidProperty(solid));
			break;
		default:
			throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to bridges");
		}
	}

	@Override
	public void clearAllContainedCheckResults() {
		super.clearAllContainedCheckResults();
		for (BoundarySurface bs : boundarySurfaces) {
			bs.clearAllContainedCheckResults();
		}
	}

	@Override
	public void collectContainedErrors(List<CheckError> errors) {
		super.collectContainedErrors(errors);
		for (BoundarySurface bs : boundarySurfaces) {
			bs.collectContainedErrors(errors);
		}
	}

	@Override
	public boolean containsAnyError() {
		boolean hasError = super.containsAnyError();
		if (hasError) {
			return true;
		}
		for (BoundarySurface bs : boundarySurfaces) {
			if (bs.containsAnyError()) {
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean containsError(CheckId checkIdentifier) {
		boolean hasError = super.containsError(checkIdentifier);
		if (hasError) {
			return true;
		}
		for (BoundarySurface bs : boundarySurfaces) {
			if (bs.containsError(checkIdentifier)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void accept(Check c) {
		super.accept(c);
		if (c.canExecute(this)) {
			c.check(this);
		}
	}

	/**
	 * Getter method for type
	 * 
	 * @return the type
	 */
	public BridgeType getType() {
		return type;
	}

	/**
	 * Setter Method for type
	 * 
	 * @param type the type to set
	 */
	public void setType(BridgeType type) {
		this.type = type;
	}

	public void addBoundarySurface(BoundarySurface bs) {
		boundarySurfaces.add(bs);
		bs.setParent(this);
	}

	@Override
	public AbstractBridge getGmlObject() {
		return ab;
	}

	@Override
	public void unsetGmlGeometries() {
		ab.unsetLod1MultiSurface();
		ab.unsetLod2MultiSurface();
		ab.unsetLod3MultiSurface();
		ab.unsetLod4MultiSurface();
		ab.unsetLod1Solid();
		ab.unsetLod2Solid();
		ab.unsetLod3Solid();
		ab.unsetLod4Solid();
	}

	public void setGmlObject(AbstractBridge aBridge) {
		ab = aBridge;
	}

	@Override
	public String toString() {
		return "BridgeObject [type=" + type + ", id=" + getGmlId() + "]";
	}

}
