/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.util.Set;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.math.Vector3d;

/**
 * 
 * @author Matthias Betz
 *
 */
public class VertexTest {

	@Test
	public void testFromVector() {
		Vector3d vec = new Vector3d(1, 2, 3);
		Vertex v = new Vertex(vec);
		assertEquals(1, v.getX(), 0.0000001);
		assertEquals(2, v.getY(), 0.0000001);
		assertEquals(3, v.getZ(), 0.0000001);
	}
	
	@Test
	public void testAddNeightbor() {
		Vertex v = new Vertex(0, 0, 0);
		Vertex neigh = new Vertex(1, 0, 0);
		v.addNeighbor(neigh);
		assertSame(neigh, v.getNeighbors().get(0));
		
		Vertex neigh2 = new Vertex(2, 0, 0);
		v.addNeighbor(neigh2);
		assertSame(neigh2, v.getNeighbors().get(1));
	}
	
	@Test
	public void testEqualsWidthNeighbor() {
		Vertex v = new Vertex(0, 0, 0);
		Vertex neigh = new Vertex(1, 0, 0);
		v.addNeighbor(neigh);
		
		Vertex other = new Vertex(1, 0, 0);
		assertTrue(v.equalsWithNeighbors(other));
		
		Vertex other2 = new Vertex(2, 0, 0);
		assertFalse(v.equalsWithNeighbors(other2));
	}
	
	@Test
	public void testNoNeighbors() {
		Vertex v = new Vertex(1, 2, 3);
		assertTrue(v.getNeighbors().isEmpty());
	}
	
	@Test
	public void testGetAdjacentRings() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Set<LinearRing> adjacentRings = v0.getAdjacentRings(geom);
		assertEquals(1, adjacentRings.size());
		assertTrue(adjacentRings.contains(lr));		
	}
	
	@Test
	public void testGetAdjacentRingsWithNeighbor() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Polygon p2 = new ConcretePolygon();
		geom.addPolygon(p2);
		
		LinearRing lr2 = new LinearRing(LinearRingType.EXTERIOR);
		p2.setExteriorRing(lr2);
		
		Vertex v1 = new Vertex(0, 0.00000001, 0);
		lr2.addVertex(v1);
		
		v0.addNeighbor(v1);
		v1.addNeighbor(v0);
		
		assertSame(v0, v1.getNeighbors().get(0));
		assertSame(v1, v0.getNeighbors().get(0));
		
		Set<LinearRing> adjacentRings = v0.getAdjacentRings(geom);
		assertEquals(2, adjacentRings.size());
		assertTrue(adjacentRings.contains(lr));		
		assertTrue(adjacentRings.contains(lr2));		

		adjacentRings = v1.getAdjacentRings(geom);
		assertEquals(2, adjacentRings.size());
		assertTrue(adjacentRings.contains(lr));		
		assertTrue(adjacentRings.contains(lr2));		

	}
	
	@Test(expected = IllegalStateException.class)
	public void testGetAdjacentRingsGeometryNotContainingVertex() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		
		Polygon p = new ConcretePolygon();
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		v0.getAdjacentRings(geom);
	}

}
