/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.CheckType;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.error.RingNotClosedError;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * CP_CLOSE checks the closeness a linear ring. A linear ring must end where it
 * started.
 * 
 * @author Matthias Betz - 12bema1bif@hft-stuttgart.de
 * 
 */
public class CloseCheck extends Check {

	private static final List<CheckId> dependencies;
	private static final List<Class<? extends Checkable>> applicableToClasses;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		dependencies = Collections.unmodifiableList(deps);
		
		ArrayList<Class<? extends Checkable>> classes = new ArrayList<>();
		classes.add(LinearRing.class);
		applicableToClasses = Collections.unmodifiableList(classes);
	}
	
	public CloseCheck() {
		super(CheckId.C_GE_R_NOT_CLOSED);
	}
	
	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}
	
	@Override
	public List<Class<? extends Checkable>> getApplicableToClasses() {
		return applicableToClasses;
	}

	@Override
	public void check(LinearRing lr) {
		Vertex first = lr.getVertices().get(0);
		Vertex last = lr.getVertices().get(lr.getVertices().size() - 1);
		CheckResult cr;
		
		if (!first.equalsWithNeighbors(last)) {
			CheckError err = new RingNotClosedError(lr);
			cr = new CheckResult(this, ResultStatus.ERROR, err);
		} else {
			cr = new CheckResult(this, ResultStatus.OK, null);
		}
		lr.addCheckResult(cr);
	}
	
	@Override
	public CheckType getType() {
		return CheckType.GEOMETRY;
	}

	@Override
	public Check createNewInstance() {
		return new CloseCheck();
	}
}
