/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.CheckType;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.check.DefaultParameter;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.Unit;
import de.hft.stuttgart.citydoctor2.check.error.NullAreaError;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Line3d;
import de.hft.stuttgart.citydoctor2.math.OrthogonalRegressionLine;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.tesselation.JoglTesselator;
import de.hft.stuttgart.citydoctor2.tesselation.TesselatedRing;

public class NullAreaCheck extends Check {
	
	private static final String DELTA_NAME = "delta";
	
	private static final List<CheckId> dependencies;
	private static final List<Class<? extends Checkable>> applicableToClasses;
	private static final List<DefaultParameter> defaultParameters;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		deps.add(CheckId.C_GE_R_DUPLICATE_POINT);
		dependencies = Collections.unmodifiableList(deps);

		ArrayList<Class<? extends Checkable>> classes = new ArrayList<>();
		classes.add(LinearRing.class);
		applicableToClasses = Collections.unmodifiableList(classes);
		
		ArrayList<DefaultParameter> defParameters = new ArrayList<>();
		defParameters.add(new DefaultParameter(DELTA_NAME, "0.0001", Unit.SQUARE_METER));
		defaultParameters = Collections.unmodifiableList(defParameters);
	}
	
	private double delta = 0.0001d;
	
	public NullAreaCheck() {
		super(CheckId.NULL_AREA);
	}

	
	@Override
	public void init(Map<String, String> parameters, ParserConfiguration config) {
		if (parameters.containsKey(DELTA_NAME)) {
			delta = Double.parseDouble(parameters.get(DELTA_NAME));
		}
	}
	
	@Override
	public List<DefaultParameter> getDefaultParameter() {
		return defaultParameters;
	}
	
	@Override
	public void check(LinearRing lr) {
		boolean areaIsNull = checkWithDistanceToRegressionLine(lr, delta);
		
		CheckResult res;
		if (areaIsNull) {
			// degenerate ring
			res = new CheckResult(this, ResultStatus.ERROR, new NullAreaError(lr));
		} else {
			res = new CheckResult(this, ResultStatus.OK, null);
		}
		lr.addCheckResult(res);
	}
	
	private boolean checkWithDistanceToRegressionLine(LinearRing lr, double delta) {
		Line3d reg = OrthogonalRegressionLine.getRegressionLine(lr.getVertices());
		for (Vertex v : lr.getVertices()) {
			if (reg.distanceToPoint(v) >= delta) {
				return false;
			}
		}
		return true;
	}

	@SuppressWarnings("unused")
	private boolean checkWithTesselation(LinearRing lr) {
		TesselatedRing tr = JoglTesselator.tesselateRing(lr);
		double area = 0D;
		for (Triangle3d t : tr.getTriangles()) {
			area = area + t.getArea();
		}
		return area < 0.0001;
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}

	@Override
	public List<Class<? extends Checkable>> getApplicableToClasses() {
		return applicableToClasses;
	}
	

	@Override
	public CheckType getType() {
		return CheckType.GEOMETRY;
	}


	@Override
	public Check createNewInstance() {
		return new NullAreaCheck();
	}

}
