/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.CheckType;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.check.GeometrySelfIntersection;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.error.SolidSelfIntError;
import de.hft.stuttgart.citydoctor2.checks.util.SelfIntersectionUtil;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.utils.PolygonIntersection;

/**
 * Check for self intersecting solids
 * 
 * @author Matthias Betz
 *
 */
public class SolidSelfIntCheck extends Check {

	private static final List<CheckId> dependencies;
	private static final List<Class<? extends Checkable>> applicableToClasses;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		dependencies = Collections.unmodifiableList(deps);

		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		deps.add(CheckId.C_GE_R_DUPLICATE_POINT);
		deps.add(CheckId.C_GE_R_SELF_INTERSECTION);
		deps.add(CheckId.C_GE_P_HOLE_OUTSIDE);
		deps.add(CheckId.C_GE_P_INNER_RINGS_NESTED);
		deps.add(CheckId.C_GE_P_INTERIOR_DISCONNECTED);
		deps.add(CheckId.C_GE_P_INTERSECTING_RINGS);
		deps.add(CheckId.C_GE_P_ORIENTATION_RINGS_SAME);
		deps.add(CheckId.C_GE_P_NON_PLANAR);
		deps.add(CheckId.C_GE_S_TOO_FEW_POLYGONS);
		deps.add(CheckId.C_GE_S_MULTIPLE_CONNECTED_COMPONENTS);
		deps.add(CheckId.C_GE_S_NON_MANIFOLD_EDGE);
		deps.add(CheckId.C_GE_S_NON_MANIFOLD_VERTEX);
		deps.add(CheckId.C_GE_S_POLYGON_WRONG_ORIENTATION);

		ArrayList<Class<? extends Checkable>> classes = new ArrayList<>();
		classes.add(Geometry.class);
		applicableToClasses = Collections.unmodifiableList(classes);
	}

	public SolidSelfIntCheck() {
		super(CheckId.C_GE_S_SELF_INTERSECTION);
	}

	@Override
	public void check(Geometry g) {
		if (g.getType() != GeometryType.SOLID && g.getType() != GeometryType.COMPOSITE_SURFACE) {
			return;
		}
		CheckResult cr;
		List<PolygonIntersection> intersections = SelfIntersectionUtil.doesSolidSelfIntersect2(g);
		if (intersections.isEmpty()) {
			cr = new CheckResult(this, ResultStatus.OK, null);
		} else {
			CheckError e = new SolidSelfIntError(g, intersections);
			cr = new CheckResult(this, ResultStatus.ERROR, e);
		}
		g.addCheckResult(cr);
	}

	@SuppressWarnings("unused")
	private CheckResult oldIntersectionAlgorithm(Geometry g) {
		CheckResult cr;
		GeometrySelfIntersection intersect = SelfIntersectionUtil.doesSolidSelfIntersect(g);
		if (intersect != null) {
			cr = new CheckResult(this, ResultStatus.ERROR, null);
		} else {
			cr = new CheckResult(this, ResultStatus.OK, null);
		}
		return cr;
	}

	@Override
	public List<Class<? extends Checkable>> getApplicableToClasses() {
		return applicableToClasses;
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}

	@Override
	public CheckType getType() {
		return CheckType.GEOMETRY;
	}

	@Override
	public Check createNewInstance() {
		return new SolidSelfIntCheck();
	}
}
