/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.semantics;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.CheckType;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.checks.util.UnfragmentedCheck;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurfaceType;

/**
 * Checks whether all polygons are in the same plane of a ground surface
 * 
 * @author Matthias Betz
 *
 */
public class GroundSurfaceUnfragmented extends Check {

	private static final List<CheckId> dependencies;
	private static final List<Class<? extends Checkable>> applicableToClasses;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		dependencies = Collections.unmodifiableList(deps);

		ArrayList<Class<? extends Checkable>> classes = new ArrayList<>();
		classes.add(BoundarySurface.class);
		applicableToClasses = Collections.unmodifiableList(classes);
	}

	public GroundSurfaceUnfragmented() {
		super(CheckId.C_SEM_BS_GROUND_NOT_FRAGMENTED);
	}

	@Override
	public void check(BoundarySurface bs) {
		// only use ground surfaces
		if (bs.getType() != BoundarySurfaceType.GROUND) {
			return;
		}
		// only use lod1 and lod2 polygons
		CheckResult cr = UnfragmentedCheck.checkForFragmentedBoundarySurfaces(this, bs, 0.1);
		if (cr == null) {
			// no error found
			cr = new CheckResult(this, ResultStatus.OK, null);
		}
		bs.addCheckResult(cr);
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}

	@Override
	public List<Class<? extends Checkable>> getApplicableToClasses() {
		return applicableToClasses;
	}

	@Override
	public CheckType getType() {
		return CheckType.SEMANTIC;
	}

	@Override
	public Check createNewInstance() {
		return new GroundSurfaceUnfragmented();
	}

}
