/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertSame;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * 
 * @author Matthias Betz
 *
 */
public class DuplicatePointsCheckTest {

	@Test
	public void testCheckOK() {
		Polygon p = new ConcretePolygon();
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(1, 0, 0);
		Vertex v2 = new Vertex(1, 1, 0);
		
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v0);
		
		DuplicatePointsCheck dpc = new DuplicatePointsCheck();
		dpc.check(lr);
		
		CheckResult cr = lr.getCheckResult(dpc);
		assertSame(ResultStatus.OK, cr.getResultStatus());
	}
	
	@Test
	public void testCheckConsecutivePointSame() {
		Polygon p = new ConcretePolygon();
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(10, 0, 0);
		Vertex v2 = new Vertex(10, 10, 0);
		
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v2);
		lr.addVertex(v0);
		
		DuplicatePointsCheck dpc = new DuplicatePointsCheck();
		dpc.check(lr);
		
		CheckResult cr = lr.getCheckResult(dpc);
		assertSame(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_CONSECUTIVE_POINTS_SAME, cr.getError().getErrorId());
	}
	
	@Test
	public void testCheckSelfInt() {
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD1);
		Vertex v0 = new Vertex(1, 1, 1);
		Vertex v1 = new Vertex(2, 2, 2);
		Vertex v2 = new Vertex(3, 3, 3);
	
		Polygon p = new ConcretePolygon();
		p.setParent(geom);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.setParent(p);
		
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v1);
		lr.addVertex(v0);
		geom.updateEdges();
		DuplicatePointsCheck dpc = new DuplicatePointsCheck();
		dpc.check(lr);
		
		CheckResult cr = lr.getCheckResult(dpc);
		assertSame(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
	}

}
