/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.geometry;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * 
 * @author Matthias Betz
 *
 */
public class HoleOutsideCheckTest {

	@Test
	public void testHoleOutside() {
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD1);
		
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		// create exterior ring
		LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(ext);
		
		Vertex firstVertex = addVertex(0, 0, 0, ext, geom);
		addVertex(3, 0, 0, ext, geom);
		addVertex(3, 3, 0, ext, geom);
		addVertex(0, 3, 0, ext, geom);
		ext.addVertex(firstVertex);
		
		// create interior ring outside exterior ring
		LinearRing inter = new LinearRing(LinearRingType.INTERIOR);
		p.addInteriorRing(inter);
		inter.setParent(p);
		
		Vertex firstIntVertex = addVertex(4, 0, 0, inter, geom);
		addVertex(5, 0, 0, inter, geom);
		addVertex(5, 1, 0, inter, geom);
		addVertex(4, 1, 0, inter, geom);
		inter.addVertex(firstIntVertex);
		
		HoleOutsideCheck c = new HoleOutsideCheck();
		c.check(p);

		CheckResult cr = p.getCheckResult(c);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_P_HOLE_OUTSIDE, cr.getError().getErrorId());
	}
	
	@Test
	public void testPolygonGood() {
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD1);
		
		Polygon poly = new ConcretePolygon();
		geom.addPolygon(poly);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		poly.setExteriorRing(lr);
		
		Vertex v0 = new Vertex(0, 0, 0);
		lr.addVertex(v0);
		
		Vertex v1 = new Vertex(10, 0, 0);
		lr.addVertex(v1);

		Vertex v2 = new Vertex(10, 10, 0);
		lr.addVertex(v2);

		Vertex v3 = new Vertex(0, 10, 0);
		lr.addVertex(v3);
		
		lr.addVertex(v0);
		
		LinearRing interior = new LinearRing(LinearRingType.INTERIOR);
		poly.addInteriorRing(interior);
		
		Vertex v4 = new Vertex(8, 2, 0);
		interior.addVertex(v4);
		
		Vertex v5 = new Vertex(8, 4, 0);
		interior.addVertex(v5);

		Vertex v6 = new Vertex(9, 4, 0);
		interior.addVertex(v6);

		Vertex v7 = new Vertex(9, 2, 0);
		interior.addVertex(v7);
		interior.addVertex(v4);

		HoleOutsideCheck c = new HoleOutsideCheck();
		c.check(poly);
		
		assertEquals(ResultStatus.OK, poly.getCheckResult(c).getResultStatus());
	}
	
	private Vertex addVertex(double x, double y, double z, LinearRing lr, Geometry geom) {
		Vertex v = new Vertex(x, y, z);
		lr.addVertex(v);
		return v;
	}

}
