/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.checks.semantics;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.util.Collections;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurfaceType;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;

/**
 * 
 * @author Matthias Betz
 *
 */
public class IsWallCheckTest {

	@Test
	public void testIsWallWarning() {
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(1, 0, 0);
		Vertex v2 = new Vertex(1, 1, 0);
		Vertex v3 = new Vertex(0, 1, 0);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v3);
		lr.addVertex(v0);
		
		Polygon p = new ConcretePolygon();
		p.setExteriorRing(lr);
		
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		geom.addPolygon(p);
		
		BoundarySurface bs = new BoundarySurface(null);
		bs.setType(BoundarySurfaceType.WALL);
		bs.addGeometry(geom);
		
		IsWallCheck check = new IsWallCheck();
		ParserConfiguration config = new ParserConfiguration(8, false);
		check.init(Collections.emptyMap(), config);
		check.check(bs);
		CheckResult checkResult = bs.getCheckResult(check);
		assertNotNull(checkResult);
		assertEquals(ResultStatus.ERROR, checkResult.getResultStatus());
	}
	
	@Test
	public void testNotWall() {
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(1, 0, 0);
		Vertex v2 = new Vertex(1, 1, 0);
		Vertex v3 = new Vertex(0, 1, 0);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v3);
		lr.addVertex(v0);
		
		Polygon p = new ConcretePolygon();
		p.setExteriorRing(lr);
		
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		geom.addPolygon(p);
		
		BoundarySurface bs = new BoundarySurface(null);
		bs.setType(BoundarySurfaceType.CLOSURE);
		bs.addGeometry(geom);
		
		IsWallCheck check = new IsWallCheck();
		ParserConfiguration config = new ParserConfiguration(8, false);
		check.init(Collections.emptyMap(), config);
		check.check(bs);
		CheckResult checkResult = bs.getCheckResult(check);
		assertNull(checkResult);
	}
	
	@Test
	public void testIsWallWarning2() {
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(0, 1, 0);
		Vertex v2 = new Vertex(1, 1, 0);
		Vertex v3 = new Vertex(1, 0, 0);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v3);
		lr.addVertex(v0);
		
		Polygon p = new ConcretePolygon();
		p.setExteriorRing(lr);
		
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		geom.addPolygon(p);
		
		BoundarySurface bs = new BoundarySurface(null);
		bs.setType(BoundarySurfaceType.WALL);
		bs.addGeometry(geom);
		
		IsWallCheck check = new IsWallCheck();
		ParserConfiguration config = new ParserConfiguration(8, false);
		check.init(Collections.emptyMap(), config);
		check.check(bs);
		CheckResult checkResult = bs.getCheckResult(check);
		assertNotNull(checkResult);
		assertEquals(ResultStatus.ERROR, checkResult.getResultStatus());
	}
	
	@Test
	public void testIsWallOk() {
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(1, 0, 0);
		Vertex v2 = new Vertex(1, 0, 1);
		Vertex v3 = new Vertex(0, 0, 1);
		
		LinearRing lr = new LinearRing(LinearRingType.EXTERIOR);
		lr.addVertex(v0);
		lr.addVertex(v1);
		lr.addVertex(v2);
		lr.addVertex(v3);
		lr.addVertex(v0);
		
		Polygon p = new ConcretePolygon();
		p.setExteriorRing(lr);
		
		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		geom.addPolygon(p);
		
		BoundarySurface bs = new BoundarySurface(null);
		bs.setType(BoundarySurfaceType.WALL);
		bs.addGeometry(geom);
		
		IsWallCheck check = new IsWallCheck();
		ParserConfiguration config = new ParserConfiguration(8, false);
		check.init(Collections.emptyMap(), config);
		check.check(bs);
		CheckResult checkResult = bs.getCheckResult(check);
		assertNotNull(checkResult);
		assertEquals(ResultStatus.OK, checkResult.getResultStatus());
	}

}
