/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.systemtest;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

/**
 * 
 * @author Matthias Betz
 *
 */
public class RingSelfIntersectionTest {
	
	@Test
	public void testSelfIntPoints1() throws CityGmlParseException, IOException, InvalidGmlFileException {
		ParserConfiguration config = new ParserConfiguration(8, false);
		CityDoctorModel m = CityGmlParser.parseCityGmlFile("src/test/resources/QA-Ex-C-LR-4.gml", config);
		Checker c = new Checker(m);
		c.runChecks();

		
		Polygon p1 = TestUtil.getPolygonById("p_w_2", m);
		LinearRing r1 = p1.getExteriorRing();
		CheckResult cr1 = r1.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
		assertEquals(ResultStatus.ERROR, cr1.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr1.getError().getErrorId());
		
		Polygon p2 = TestUtil.getPolygonById("p_w_4", m);
		LinearRing r2 = p2.getExteriorRing();
		assertEquals(ResultStatus.OK, r2.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION).getResultStatus());
	}
	
	@Test
	public void testSelfIntPoints2() throws CityGmlParseException, IOException, InvalidGmlFileException {
		ParserConfiguration config = new ParserConfiguration(8, false);
		CityDoctorModel m = CityGmlParser.parseCityGmlFile("src/test/resources/QA-Ex-C-LR-4GK.gml", config);
		Checker c = new Checker(m);
		c.runChecks();
		
		Polygon p1 = TestUtil.getPolygonById("p_w_2", m);
		LinearRing r1 = p1.getExteriorRing();
		CheckResult cr1 = r1.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
		assertEquals(ResultStatus.ERROR, cr1.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr1.getError().getErrorId());
		
		Polygon p2 = TestUtil.getPolygonById("p_w_4", m);
		LinearRing r2 = p2.getExteriorRing();
		assertEquals(ResultStatus.OK, r2.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION).getResultStatus());
	}
	
	@Test
	public void testSelfIntPoints3() throws CityGmlParseException, IOException, InvalidGmlFileException {
		ParserConfiguration config = new ParserConfiguration(8, false);
		CityDoctorModel m = CityGmlParser.parseCityGmlFile("src/test/resources/QA-Ex-C-Solid-1.gml", config);
		Checker c = new Checker(m);
		c.runChecks();
		
		Polygon p1 = TestUtil.getPolygonById("p_w_2", m);
		LinearRing r1 = p1.getExteriorRing();
		CheckResult cr1 = r1.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
		assertEquals(ResultStatus.OK, cr1.getResultStatus());
		
		Polygon p2 = TestUtil.getPolygonById("p_w_4", m);
		LinearRing r2 = p2.getExteriorRing();
		assertEquals(ResultStatus.OK, r2.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION).getResultStatus());
	}
	
	@Test
	public void testSelfIntPoints4() throws CityGmlParseException, IOException, InvalidGmlFileException {
		CityDoctorModel c = TestUtil.loadAndCheckCityModel("src/test/resources/SimpleSolid_SrefBS.gml");
		for (Polygon p : c.getBuildings().get(0).getGeometries().get(0).getPolygons()) {
			LinearRing r = p.getExteriorRing();
			CheckResult cr = r.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
			assertEquals(ResultStatus.OK, cr.getResultStatus());
		}
	}
	
	@Test
	public void testSelfIntPoints5() throws CityGmlParseException, IOException, InvalidGmlFileException {
		CityDoctorModel c = TestUtil.loadAndCheckCityModel("src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0004-T0004.gml");
		Polygon p = TestUtil.getPolygonById("_Simple_BD.1_PG.6", c);
		LinearRing r = p.getExteriorRing();
		CheckResult cr = r.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
	}
	
	@Test
	public void testSelfIntPoints6() throws CityGmlParseException, IOException, InvalidGmlFileException {
		Map<String, String> param = new HashMap<>();
		param.put("minVertexDistance", "0.1");
		Map<CheckId, Map<String, String>> params = new HashMap<>();
		params.put(CheckId.C_GE_R_SELF_INTERSECTION, param);
		CityDoctorModel c = TestUtil.loadAndCheckCityModel("src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0004-T0005.gml", params);
		Polygon p = TestUtil.getPolygonById("_Simple_BD.1_PG.2", c);
		LinearRing r = p.getExteriorRing();
		CheckResult cr = r.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
	}
	
	@Test
	public void testSelfIntPoints7() throws CityGmlParseException, IOException, InvalidGmlFileException {
		CityDoctorModel c = TestUtil.loadAndCheckCityModel("src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0004-T0005.gml", 2);
		for (Polygon p : c.getBuildings().get(0).getGeometries().get(0).getPolygons()) {
			LinearRing r = p.getExteriorRing();
			CheckResult cr = r.getCheckResult(CheckId.C_GE_R_SELF_INTERSECTION);
			assertEquals(ResultStatus.OK, cr.getResultStatus());
		}
	}

}
